#include <stdlib.h>
#include <math.h>
#include "terrain.h"

#define PI	3.141592654
#define TWOPI	6.283185308

// the Fourier transform code from Numerical Recipes
#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr

void fourn(float data[], unsigned long nn[], int ndim, int isign)
{
        int idim;
        unsigned long i1,i2,i3,i2rev,i3rev,ip1,ip2,ip3,ifp1,ifp2;
        unsigned long ibit,k1,k2,n,nprev,nrem,ntot;
        float tempi,tempr;
        double theta,wi,wpi,wpr,wr,wtemp;

        for (ntot=1,idim=1;idim<=ndim;idim++)
                ntot *= nn[idim];
        nprev=1;
        for (idim=ndim;idim>=1;idim--) {
                n=nn[idim];
                nrem=ntot/(n*nprev);
                ip1=nprev << 1;
                ip2=ip1*n;
                ip3=ip2*nrem;
                i2rev=1;
                for (i2=1;i2<=ip2;i2+=ip1) {
                        if (i2 < i2rev) {
                                for (i1=i2;i1<=i2+ip1-2;i1+=2) {
                                        for (i3=i1;i3<=ip3;i3+=ip2) {
                                                i3rev=i2rev+i3-i2;
                                                SWAP(data[i3],data[i3rev]);
                                                SWAP(data[i3+1],data[i3rev+1]);
                                        }
                                }
                        }
                        ibit=ip2 >> 1;
                        while (ibit >= ip1 && i2rev > ibit) {
                                i2rev -= ibit;
                                ibit >>= 1;
                        }
                        i2rev += ibit;
                }
                ifp1=ip1;
                while (ifp1 < ip2) {
                        ifp2=ifp1 << 1;
                        theta=isign*6.28318530717959/(ifp2/ip1);
                        wtemp=sin(0.5*theta);
                        wpr = -2.0*wtemp*wtemp;
                        wpi=sin(theta);
                        wr=1.0;
                        wi=0.0;
                        for (i3=1;i3<=ifp1;i3+=ip1) {
                                for (i1=i3;i1<=i3+ip1-2;i1+=2) {
                                        for (i2=i1;i2<=ip3;i2+=ifp2) {
                                                k1=i2;
                                                k2=k1+ifp1;
                                                tempr=(float)wr*data[k2]-(float)wi*data[k2+1];
                                                tempi=(float)wr*data[k2+1]+(float)wi*data[k2];
                                                data[k2]=data[k1]-tempr;
                                                data[k2+1]=data[k1+1]-tempi;
                                                data[k1] += tempr;
                                                data[k1+1] += tempi;
                                        }
                                }
                                wr=(wtemp=wr)*wpr-wi*wpi+wr;
                                wi=wi*wpr+wtemp*wpi+wi;
                        }
                        ifp1=ifp2;
                }
                nprev *= n;
        }
}
#undef SWAP


// convolves source with conv, storing the result in conv
int trn_convolve(terrain_t *source, terrain_t *conv)
{
	float *sc, *cv;
	float *a, *b, *c, *d, i, j;
	int l;
	int f, g;
	unsigned long nn[2];

	// are the two maps conformable?
	if (source->w != conv->w || source->h != conv->h)
		return 1;

	// are the maps power-of-two dimensioned?
	if (((source->w - 1) | source->w) >> 1 != (source->w - 1) ||
	    ((source->h - 1) | source->h) >> 1 != (source->h - 1))
		return 2;

	l = 2 * source->w * source->h;
	nn[0] = source->w;
	nn[1] = source->h;

	sc = (float*)malloc(l * sizeof(float));
	if (!sc) return 3;
	cv = (float*)malloc(l * sizeof(float));
	if (!cv)
	{
		free(sc);
		return 3;
	}

	memset(sc, 0, l * sizeof(float));
	memset(cv, 0, l * sizeof(float));

	// copy data in
/*	for (a = source->terrain, b = sc, c = conv->terrain, d = cv;
		b < sc + l; a++, b += 2, c++, d += 2)
	{
		*b = *a;
		*d = *c;
	}
*/
	for (f = 0; f < source->w; f++)
	for (g = 0; g < source->h; g++)
	{
		*(sc + 2 * (f + g * source->w)) = source->terrain[f + g * source->w];
		*(cv + 2 * (f + g * source->w)) = conv->terrain[(f + source->w / 2) % source->w + ((g + source->h / 2) % source->h) * source->w];
	}

	// go through the dark portal into k-space
	fourn(sc - 1, nn - 1, 2, 1);
	fourn(cv - 1, nn - 1, 2, 1);

	// do the convolution
	for (a = sc, b = sc + 1, c = cv, d = cv + 1; a < sc + l ;
		a += 2, b += 2, c += 2, d += 2)
	{
		i = *a * *c - *b * *d;
		j = *a * *d + *b * *c;

		*a = i;
		*b = j;
	}

	// untransform
	fourn(sc - 1, nn - 1, 2, -1);

	// copy back
	for (a = source->terrain, b = sc; b < sc + l; a++, b += 2)
	{
		// crazy normalisation factor from fft
		*a = *b / (source->w * source->w * 4 * PI);
	}

	free(sc);
	free(cv);

	return 0;
}

// add some gaussian noise in a terrain
//
// noise has amplitude mag * exp(-(k / dev)^2) and random phase
int trn_gaussiannoise(terrain_t *t, float mag, float dev)
{
	float *d, kk, dd, ph, *p, *q;
	float smallfurrymammal;
	int l;
	unsigned long nn[2];
	int i, j, a, b;

	if (!t) return 1;
	
	// is the map power-of-two dimensioned?
	if (((t->w - 1) | t->w) >> 1 != (t->w - 1) ||
	    ((t->h - 1) | t->h) >> 1 != (t->h - 1))
		return 2;

	nn[0] = t->w;
	nn[1] = t->h;

	l = t->w * t->h;

	d = (float*)malloc(2 * l * sizeof(float));
	if (!d)
		return 1;

	dd = dev * dev;// * (t->w * t->w);

	fprintf(stderr,"Giraffe %e %f %e\n",dd,dev,mag);

	mag /= sqrt(t->w * t->h) / PI;
	
	// fill d with complex numbers of arbitrary phase and gaussian
	// amplitude
	for (i = 1; i < t->w / 2; i++) // start from 1, so no DC component
	for (j = 1; j < t->h / 2; j++)
	{
		// k^2 is given by i^2 pi^2 scale^2 / w^2 + j^2...
		kk = i * i * PI * PI * t->xs * t->xs / (t->w * t->w)
			+ j * j * PI * PI * t->ys * t->ys / (t->h * t->h);

		// fill in the 4 corners of k-space differently
		// top left
		ph = TWOPI * ((float)rand() / (float)RAND_MAX);

		if ((i==10) && (j==10)) {
			fprintf(stderr,"Soup %f %f %e %f %f\n",t->xs,t->ys,dd,kk,ph);
		}
		a = i;
		b = j;

		smallfurrymammal = (float)rand() / (float)RAND_MAX;

		*(d + (a + b * t->h) * 2)     = exp(-kk / dd) * cos(ph) * mag * smallfurrymammal;
		*(d + (a + b * t->h) * 2 + 1) = exp(-kk / dd) * sin(ph) * mag * smallfurrymammal;

//printf("%f %f\n", *(d + (a + b * t->h) * 2),*(d + (a + b * t->h) * 2 + 1));
		
		// top right
		ph = TWOPI * ((float)rand() / (float)RAND_MAX);

		a = t->w - i;
		b = j;

		*(d + (a + b * t->h) * 2)     = exp(-kk / dd) * cos(ph) * mag * smallfurrymammal;
		*(d + (a + b * t->h) * 2 + 1) = exp(-kk / dd) * sin(ph) * mag * smallfurrymammal;

		// bottom left
		ph = TWOPI * ((float)rand() / (float)RAND_MAX);

		a = i;
		b = t->h - j;

		*(d + (a + b * t->h) * 2)     = exp(-kk / dd) * cos(ph) * mag * smallfurrymammal;
		*(d + (a + b * t->h) * 2 + 1) = exp(-kk / dd) * sin(ph) * mag * smallfurrymammal;

		// bottom right
		ph = TWOPI * ((float)rand() / (float)RAND_MAX);

		a = t->w - i;
		b = t->h - j;

		*(d + (a + b * t->h) * 2)     = exp(-kk / dd) * cos(ph) * mag * smallfurrymammal;
		*(d + (a + b * t->h) * 2 + 1) = exp(-kk / dd) * sin(ph) * mag * smallfurrymammal;
	}

	// inverse FT of k
	fourn(d - 1, nn - 1, 2, -1);
//	memset(d, 0, l * sizeof(float) * 2);
	fprintf(stderr, "fish\n");

	// copy stuff (real part only) into terrain
	for (p = t->terrain, q = d; p < t->terrain + l; p++, q += 2)
		*p = *q; // / (t->w);
	
	free(d);

	return 0;
}
